/*-
 * Copyright (C)2006..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2006..2021 @BABOLO http://www.babolo.ru/\n"
#ident "@(#) All rights reserved.\n"
#ident "@(#) $Id: getord.c,v 1.19 2021/06/06 17:37:13 babolo Exp $\n"

#define BLIN_COMPAT  4
#define Bpars_COMPAT VMAJOR

#include <sys/types.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include "parser.h"

static u_int32_t ofm[] = {0x00000000, 0x000000FF, 0x0000FFFF, 0x00FFFFFF, 0xFFFFFFFF};

static u_int32_t
/*****************************************************************************************************/
reconstruct(const babolo_lexor *lext, size_t *lexind) {                                          /****
 *****************************************************************************************************/
# define blin_internal_flags (lext->flags & BLIN_MASK)
    u_int32_t i;
    u_int32_t e;

    for (i = 0, e = 0; i < lext->szp; e += lext->lexgraf[(*lexind)++] << (8 * i++));
    ifBLIN_QX5("  %08X <= %08X ?", e, lext->szz);
    if  (e <= lext->szz) {
        if  (!(lext->flags & Bpars_AABS)) e += (lext->szp + *lexind) / (lext->szp + 1);
        ifBLIN_QX5("Y %08X", e);
    } else {
        e |= ~ofm[lext->szp];
        ifBLIN_QX5("N %08X |%08X", e, ~ofm[lext->szp]);
    }
    return(e);
# undef blin_internal_flags
}

u_int32_t
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_testchar(const babolo_lexor *lext, u_int32_t state, const char **next) {                  /****
 *****************************************************************************************************
 *****************************************************************************************************/
# define blin_internal_flags (lext->flags & BLIN_MASK)
    size_t lexind;
    size_t e = ~0;
    int    found;
    u_char nxch;
    size_t n;
    u_char c;

    ifBLIN_QX3("+ flags=%06X state=%d next=%02X(%c)", lext->flags, state, **next & 0xFF, **next);
    nxch = (lext->retable) ? lext->retable[**next & 0xFF] : **next & 0xFF;
    if  (state < lext->szt) {
        lexind = (lext->szp + 1) * state;
        found = 0;
        for (n = lext->lexgraf[lexind++] + 1; n > 0; --n) {
            c = lext->lexgraf[lexind++];
            ifBLIN_QX4("i%d %02X(%c)", (int)n, c, c);
            if  (c == nxch) {
                e = reconstruct(lext, &lexind);
                ++*next;
                found = 1;
                break;
            } else {
                lexind += lext->szp;
        }   }
        if  (!found) {
            e = reconstruct(lext, &lexind);
            if  (e < lext->szt) {
                if  (!(lext->flags & Bpars_CMID)) ++*next;
            } else {
                if  (lext->flags & Bpars_CEND) ++*next;
    }   }   }
    ifBLIN_QX3("- %d next=%02X", (int)e, nxch);
    return(e);
# undef blin_internal_flags
}

static u_int32_t
/*****************************************************************************************************/
babolo_weakchar(const babolo_lexor *lext, u_int32_t state) {                                     /****
 *****************************************************************************************************/
# define blin_internal_flags (lext->flags & BLIN_MASK)
    size_t lexind = 0;
    size_t e = ~0;
    size_t n;

    ifBLIN_QX3("+ flags=%06X state=%d", lext->flags, state);
    if  (state < lext->szt) {
        lexind = (lext->szp + 1) * state;
        n = lext->lexgraf[lexind++] + 1;
        lexind += (lext->szp + 1) * n;
        e = reconstruct(lext, &lexind);
    }
    ifBLIN_QX3("- %d [%d]", (int)e, (int)lexind);
    return(e);
# undef blin_internal_flags
}

u_int32_t
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_goword(const babolo_lexor *lext, const char **word) {                                     /****
 *****************************************************************************************************
 *****************************************************************************************************/
# define blin_internal_flags (lext->flags & BLIN_MASK)
    u_int32_t     maxcy = (lext->szt + lext->szp + 2) / (lext->szp + 1);           /* Max cicle size */
    const u_char *wcpy = NULL;
    u_int32_t     state;
    u_int32_t     curcy;
    u_int32_t     j = 0;
    u_int32_t     k;
    const u_char *v;

    ifBLIN_QX3("+ %s~", *word);
    for ( state = 0, curcy = maxcy, v = (u_char*)*word
        ; state < lext->szt
        ; state = babolo_testchar(lext, state, word), --curcy
        ) {
        if  (v != (u_char*)*word) {
            curcy = maxcy;
            v = (u_char*)*word;
        } else if (!curcy) {
            state = ~0;
            j = 0;
            ifBLIN_QX1("infinite loop %u", maxcy);
            break;
        }
        k = babolo_weakchar(lext, state);
        if  (k >= lext->szt && ~k) {
            j = ~k;
            wcpy = (u_char*)*word;
            ifBLIN_QX4("Reserve %u", j);
    }   }
    k = ~state;
    if  (!k && j) {
        k = j;
        *word = (char*)wcpy;
        ifBLIN_QX4("Reserved %u", k);
    }
    ifBLIN_QX3("- %d", k);
    return(k);
# undef blin_internal_flags
}

u_int32_t
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_testword(const babolo_lexor *lext, const char *word) {                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
    return(babolo_goword(lext, &word));
}

void
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_treedump(const babolo_lexor *lext) {                                                      /****
 *****************************************************************************************************
 *****************************************************************************************************/
    size_t    lexind;
    u_char    c;
    u_int32_t j;
    u_int32_t k;
    u_int32_t m;
    u_int32_t e;
    u_int32_t n;

    m = 0;
    lexind = 0;
    k = 0;
    for (;;) {
        n = lext->lexgraf[lexind++] + 1;
        fprintf(stderr, "[%2u]", k);
        for (j = 0; j < n; ++j) {
            if  (lexind > (lext->szp + 1) * lext->szt) {
                fprintf(stderr, "Premature end\n");
                return;
            }
            c = lext->lexgraf[lexind++];
            if  (c >= ' ' && c < 127) fprintf(stderr, "   (%c)%02X", c, c);
              else fprintf(stderr, "      %02X", c);
            e = reconstruct(lext, &lexind);
            if  (m < e && e < lext->szt) m = e;
            if  (e < lext->szt) fprintf(stderr, "%c%2u", (e > (k + j)) ? '.' : '!', e);
              else fprintf(stderr, "%c~%u", (e > (k + j)) ? '.' : '!', ~e & ofm[lext->szp]);
        }
        e = reconstruct(lext, &lexind);
        if  (m < e && e < lext->szt) m = e;
        k += n + 1;
        if  (e < lext->szt) fprintf(stderr, "   *%c%2u\n", (e >= k) ? '.' : '!', e);
          else fprintf(stderr, "   *%c~%u\n", (e >= k) ? '.' : '!', ~e & ofm[lext->szp]);
        if  (k >= lext->szt) break;
        if  (k > m) fprintf(stderr, "No way\n");
    }
    if  (m >= lext->szt) fprintf(stderr, "Link out\n");
}
